from collections.abc import Container, Iterable, Sized
from itertools import chain
from copy import copy


class SpanError(Exception):
    pass


class Span(Container, Iterable, Sized):

    __slots__ = ["a", "b", "context_length", "cyclic", "index"]

    def __init__(self, a, b, l, cyclic=False, index=0, allow_wrap=False):
        if b is not None and a > b and not cyclic:
            raise IndexError(
                "Start {} cannot exceed end {} for linear spans".format(a, b)
            )

        self.index = index
        self.context_length = l
        self.cyclic = cyclic
        if not cyclic or not allow_wrap:
            bounds = self.bounds()
            if not bounds[0] <= a <= bounds[1]:
                raise IndexError(
                    "Start {} must be within ({}, {})".format(a, index, index + l)
                )
            if not bounds[0] <= b <= bounds[1]:
                raise IndexError(
                    "End {} must be within ({}, {})".format(b, index, index + l)
                )
        if a > l or a < 0:
            self.a = self.t(a)
        else:
            self.a = a
        if b > l or b < 0:
            self.b = self.t(b)
        else:
            self.b = b

    def bounds(self):
        """Return the bounds (end exclusive)"""
        return self.index, self.context_length + self.index

    def t(self, p):
        """Translate an index to a valid index"""
        if p >= self.bounds()[1] or p < self.bounds()[0]:
            if not self.cyclic:
                raise IndexError(
                    "Position {} outside of linear bounds {}".format(p, self.bounds())
                )
        return (p % self.context_length) + self.bounds()[0]

    def ranges(self):
        if self.cyclic and self.a > self.b:
            return [(self.a, self.bounds()[1]), (self.bounds()[0], self.b)]
        else:
            return [(self.a, self.b)]

    def new(self, a, b):
        return self.__class__(a, b, self.context_length, self.cyclic)

    def sub(self, a, b):
        if b is not None and a > b and not self.cyclic:
            raise ValueError(
                "Start {} cannot exceed end {} for linear spans".format(a, b)
            )
        copied = copy(self)
        copied.a = a
        copied.b = b
        return copied

    def same_context(self, other):
        return (
            other.context_length == self.context_length and self.cyclic == other.cyclic
        )

    def force_context(self, other):
        if not self.same_context(other):
            raise SpanError("Cannot compare with different contexts")

    def differences(self, other):
        """Return a tuple of differences between this span and the other span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return (self.sub(self.a, other.a),)
        elif other.b in self and other.a not in self:
            return (self.sub(other.b, self.b),)
        if other in self:
            return self.sub(self.a, other.a), self.sub(other.b, self.b)
        elif self in other:
            return (self.sub(self.a, self.a),)
        else:
            return ((self[:]),)

    def intersection(self, other):
        """Return the span inersection between this span and the other span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return self.sub(other.a, self.b)
        elif other.b in self and other.a not in self:
            return self.sub(self.a, other.b)
        if other in self:
            return other[:]
        elif self in other:
            return self[:]

    def consecutive(self, other):
        try:
            return self.b == other.t(other.a - 1)
        except IndexError:
            return False

    # def union(self, other):
    #     self.force_context(other)
    #     if other in self:
    #         return self[:]
    #     elif self in other:
    #         return other[:]
    #     elif other.a in self and not other.t(other.b - 1) in self:
    #         if self.a == other.b:
    #             return self.new(self.a, None)
    #         return self.new(self.a, other.b)
    #     elif other.t(other.b - 1) in self and other.a not in self:
    #         if other.a == self.b:
    #             return self.new(self.a, None)
    #         return self.new(other.a, self.b)

    #     def __le__(self, other):
    #         return self.a <= other.a

    #     def __lt__(self, other):
    #         return self.a < other.a

    #     def __gt__(self, other):
    #         return self.a > other.a

    #     def __ge__(self, other):
    #         return self.a >= other.a

    def __eq__(self, other):
        return self.same_context(other) and self.a == other.a and self.b == other.b

    def __ne__(self, other):
        return not (self.__eq__(other))

    def __contains__(self, other):
        if isinstance(other, int):
            for r in self.ranges():
                if r[0] <= other < r[1]:
                    return True
            return False
        elif issubclass(type(other), Span):
            if not self.same_context(other):
                return False
            return (
                other.a in self
                and other.t(other.b - 1) in self
                and len(other) <= len(self)
            )

    def __len__(self):
        return sum([r[1] - r[0] for r in self.ranges()])

    def __iter__(self):
        for i in chain(*[range(*x) for x in self.ranges()]):
            yield i

    def __getitem__(self, val):
        if isinstance(val, int):
            if not self.cyclic and (val >= len(self) or val < -len(self)):
                raise IndexError(
                    "Index '{}' outside of linear span with length of {}".format(
                        val, len(self)
                    )
                )
            else:
                if val < 0:
                    return self.t(val + self.b) - self.index
                else:
                    return self.t(val + self.a) - self.index
        elif issubclass(type(val), slice):
            if val.step:
                raise ValueError(
                    "{} slicing does not support step.".format(self.__class__.__name__)
                )
            if val.start is None:
                i = self.a
            else:
                i = self[val.start]

            if val.stop is None:
                j = self.b
            else:
                j = self[val.stop]
            return self.sub(i, j)
        elif isinstance(val, tuple):
            if len(val) > 2:
                raise ValueError(
                    "{} -- copying does only supports (start, stop)".format(val)
                )
            return self.new(*val)
        else:
            raise ValueError("indexing does not support {}".format(type(val)))

    def __repr__(self):
        return "<{}={} {} {}>".format(
            self.__class__.__name__,
            id(self),
            (self.a, self.b, self.context_length),
            self.cyclic,
        )

    def __str__(self):
        return "<{} {} {}>".format(
            self.__class__.__name__, (self.a, self.b, self.context_length), self.cyclic
        )



from Bio.Seq import Seq
import random
from Bio.Alphabet.IUPAC import IUPACUnambiguousDNA
from Bio.SeqFeature import FeatureLocation, CompoundLocation


def random_sequence(size, alphabet, seqclass=Seq):
    letters = [random.choice(alphabet.letters) for _ in range(size)]
    return seqclass("".join(letters))


def random_dna(size, seqclass=Seq):
    return random_sequence(size, IUPACUnambiguousDNA, seqclass)


def topology(record):
    return record.annotations.get("topology", "linear")


def is_circular(record):
    return topology(record).strip().lower() == "circular"


def make_location_parts(start, end, length, strand):
    if start > end:
        if length is None:
            raise ValueError(
                "A length must be provided to create a feature with start > end."
            )
        f1 = FeatureLocation(start, length, strand)
        f2 = FeatureLocation(1, end, strand)
        if strand == -1:
            parts = [f2, f1]
        else:
            parts = [f1, f2]
    else:
        parts = [FeatureLocation(start, end, strand=strand)]
    return parts


def make_circular_location(start, end, length, strand):
    """
    Make a location. If spans origin, a CompoundLocation is created.

    :param start: inclusive start position
    :type start: int
    :param end: exclusive end position
    :type end: int
    :param length: length of the sequence
    :type length: int
    :param strand: strand (1 or -2)
    :type strand: int
    :return:
    :rtype:
    """
    parts = make_location_parts(start, end, length, strand)
    if len(parts) > 1:
        return CompoundLocation(parts)
    else:
        return parts[0]



# __version__.py
# autogenerated by keats 0.2.24
__version__ = "0.5.4"
__name__ = "lobio"
__title__ = "lobio"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repo__ = None
__homepage__ = "https://www.github.com/jvrana/LoBio"
__description__ = "Extra utilities for BioPython"



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['lobio']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.73,<2.0']

setup_kwargs = {
    'name': 'lobio',
    'version': '0.5.4',
    'description': 'Extra utilities for BioPython',
    'long_description': '# LoBio\n\nKitchen sink utilities for BioPython\n\n## ImmutableSeqRecord\n\n',
    'author': 'Justin Vrana',
    'author_email': 'justin.vrana@gmail.com',
    'url': 'https://www.github.com/jvrana/LoBio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from abc import ABC

from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, CompoundLocation
from functools import wraps
from .region import Region
from .SeqUtils import make_circular_location, is_circular
from copy import deepcopy
from .exceptions import StrictFeatureError


def region_to_feature_location(region):
    """Convert a Region to a FeatureLocation or CompoundLocation if appropriate."""
    return make_circular_location(
        region.a, region.b, region.context_length, region.direction
    )


def feature_locations_to_regions(locations, length, cyclic):
    """Convert locations to a list of regions."""
    regions = []
    for location in locations:
        for part in location.parts:
            strand = part.strand
            regions.append(
                Region(part.start, part.end, length, cyclic=cyclic, direction=strand)
            )
    return regions


def bind_feature_to_context(feature, length, cyclic, start=None, end=None, strand=None):
    """
    Validates a SeqFeature against sequence topology and size. Modifies location features
    to match the context. If location is a Compound
    """
    locations = []
    if None in [start, end, strand]:
        if not all(x is None for x in [start, end, strand]):
            raise ValueError(
                "Arguments 'start', 'end', 'and' strand must either all be None or all be ints."
            )
        if feature.location:
            regions = feature_locations_to_regions(
                feature.location.parts, length, cyclic
            )
            locations += [region_to_feature_location(r) for r in regions]
    else:
        region = Region(start, end, length, cyclic=cyclic, direction=strand)
        locations.append(region_to_feature_location(region))
    if len(locations) > 1:
        feature.location = CompoundLocation(locations)
    elif locations:
        feature.location = locations[0]
    else:
        raise StrictFeatureError("Cannot create feature. No locations were found.")
    return feature


class ImmutableException(Exception):
    @classmethod
    def do_raise(cls, instance, prop, msg=""):
        errmsg = "Cannot set immutable property '{}'. {} is immutable.".format(
            prop, instance.__class__
        )
        if msg:
            errmsg += " " + msg
        return cls(errmsg)


class ImmutableSeqFeature(SeqFeature):
    @wraps(SeqFeature.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_seq_feature(self):
        pass

    @property
    def location(self):
        return super().location

    @location.setter
    def location(self, l):
        if self.location:
            raise ImmutableException.do_raise(self, "location")
        else:
            super().location = l


class Constants(object):

    TOPOLOGY = "topology"
    CIRCULAR = "circular"
    LINEAR = "linear"


class ImmutableSeqRecord(SeqRecord, ABC):
    """
    A SeqRecord object with strict requirements for locations of SeqFeatures.

    Features:
        1. feature locations are validated against topology and size of sequence
        2. Seq instance cannot be changed
    """

    @wraps(SeqRecord.__init__)
    def __init__(self, *args, circular=False, **kwargs):
        self._features = []
        super().__init__(*args, **kwargs)
        self.circular = circular
        for f in self.features:
            bind_feature_to_context(f)

    def _init(self):
        for f in self.features:
            bind_feature_to_context(f)

    @classmethod
    def from_seq_record(cls, record):
        return cls(
            seq=record.seq,
            id=record.id,
            name=record.name,
            description=record.description,
            dbxrefs=deepcopy(record.dbxrefs),
            features=deepcopy(record.features),
            annotations=deepcopy(record.annotations),
            letter_annotations=deepcopy(record.letter_annotations),
        )

    def to_seq_record(self):
        return SeqRecord(
            seq=self.seq,
            id=self.id,
            name=self.name,
            description=self.description,
            dbxrefs=deepcopy(self.dbxrefs),
            features=list(deepcopy(self.features)),
            annotations=deepcopy(self.annotations),
            letter_annotations=deepcopy(self.letter_annotations),
        )

    @property
    def circular(self):
        return is_circular(self)

    @circular.setter
    def circular(self, b):
        if b:
            self.annotations[Constants.TOPOLOGY] = Constants.CIRCULAR
        else:
            self.annotations[Constants.TOPOLOGY] = Constants.LINEAR

    @property
    def features(self):
        return tuple(self._features)

    def __add__(self, other):
        return self.from_seq_record(self.to_seq_record() + other.to_seq_record())

    @features.setter
    def features(self, featurelist):
        for f in featurelist:
            self.add_feature(f)

    def add_feature(self, feature, start=None, end=None, strand=None):
        bind_feature_to_context(feature, len(self), self.circular, start, end, strand)
        self._features.append(feature)

    @property
    def context(self):
        return len(self), self.circular

    @property
    def seq(self):
        return super().seq

    @seq.setter
    def seq(self, s):
        if self.seq:
            raise ImmutableException.do_raise(self, "seq")
        else:
            super().seq = s



"""
Project: jdna
File: region
Author: Justin
Date: 2/21/17

Description: Basic functionality for defining regions of linear, circularized, or reversed
regions of a sequence.

"""
from lobio.span import Span


class Direction(object):
    FORWARD = 1
    REVERSE = -1
    BOTH = 0


class Region(Span):
    __slots__ = ["name", "id", "direction"]

    FORWARD = Direction.FORWARD
    REVERSE = Direction.REVERSE
    BOTH = Direction.BOTH

    def __init__(
        self,
        start,
        end,
        length,
        cyclic=False,
        index=0,
        direction=FORWARD,
        name=None,
        id=None,
        allow_wrap=False,
    ):
        self.name = name
        self.id = id
        assert direction in [self.FORWARD, self.REVERSE, self.BOTH]
        self.direction = direction
        super().__init__(
            start, end, length, cyclic=cyclic, index=index, allow_wrap=allow_wrap
        )

    @property
    def start(self):
        if self.direction == self.REVERSE:
            return self.b
        else:
            return self.a

    @property
    def end(self):
        if self.direction == self.REVERSE:
            return self.a
        else:
            return self.b


# class Context(object):
#
#     __slots__ = ["length", "cyclic", "index"]
#
#     def __init__(self, length, cyclic=False, index=0):
#         self.length = length
#         self.cyclic = cyclic
#         self.index = index
#
#     def region(self, start, end, direction=Direction.FORWARD, name=None, id=None):
#         return Region(
#             start,
#             end,
#             self.length,
#             cyclic=self.cyclic,
#             index=self.index,
#             direction=direction,
#             name=name,
#             id=id,
#         )
